
/*
 * meta3d: powerful convertor of 3D Studio *.asc to METAFONT source code
 *
 * (C) 1999 by Daniel S\'{y}kora (e-mail: sykorad@feld.cvut.cz)
 *
 * 18th April 1999
 * version: 0.5
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public Licenses as by published
 * by the Free Software Foundation; either version 2; or (at your option)
 * any later version
 *
 * This program is distributed in the hope that it will entertaining,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 *  +--------------+
 *  |  Meta3D v0.5 |
 *  +--------------+
 *
 * code: ReDox/MovSD (a.k.a Daniel S\'{y}kora) * 1999
 *
 * Fast portable convertor of readable 3D Studio output format *.asc
 * to source code for powerful METAFONT or METAPOST renderer.
 * Easy way how to generate handsome portable and resolution
 * independent 3D wire pictures for TeX articles.
 *
 * Features:
 *
 * + free source code and platform portability
 * + full dynamic memory allocation -- no inner limitations
 * + fast and good-looking visibility rasterization
 * + redundancy optimalization of METAFONT output
 * + smart 3D Studio camera support with border clipping
 * + easy pictures maintaining in PlainTeX articles
 * + easy portability of input format
 *
 * - huge METAFONT code in large scenes
 * - non-analytical visibility calculation
 * - not so exact 3D Studio camera support
 * - quantization noise provides not so correct hide-removal results
 * - pure depth-sort painter's algorithm instead of quality BSP-tree
 *   visibility approach (will be in future)
 */

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#define MAX_INPUT_LINE_LENGTH 100

int xres=1280,
    yres=800,
    fovx=256*4,
    fovy=256*4,
    border=0,
    hide_removal=1,
    invert_normals=0;

char *hsize="10cm",
     *aspect="0.625",
     *penth="0.1mm";

typedef float MATRIX[4][4];

typedef float VECTOR[3];

typedef struct {
        int   id;
        int   used;
        float sx;
        float sy;
        float x;
        float y;
        float z;
        float tx;
        float ty;
        float tz;
        } VERTEX;

typedef struct {
        VERTEX v1,v2;
        } LINE;

typedef struct {
        VERTEX *vertex[3];
        } FACE;

typedef struct _LIST {
        LINE line;
        int visible;
        struct _LIST * next;
        } LIST;

typedef struct {
        int v1_id,v2_id;
        int visible;
        int sortkey;
        LINE line;
        LIST *cut;
        VERTEX *v1,*v2;
        } INFO;

typedef struct _TREE {
        float z;
        FACE * face;
        struct _TREE * more;
        struct _TREE * less;
        } TREE;

typedef struct {
        int    *clipping_buffer;
        int    vertex_size;
        int    face_size;
        FACE   *face;
        INFO   *info;
        TREE   *tree;
        VERTEX *vertex;
        MATRIX *camera;
        } WORLD;

int inline round(double d)
{
 int i;
 double _d;
 if (modf(d,&_d)>0.5) _d++;
 i=_d;
 return i;
}

void rasterize_line(WORLD *w, VERTEX *p1, VERTEX *p2)
{
  float x,y,x1,x2,y1,y2,dx,dy;
  int npix,ix,iy,id;

  if (p1->id>p2->id) {
     x2 = round(p1->sx); x1 = round(p2->sx);
     y2 = round(p1->sy); y1 = round(p2->sy);
     id = p2->id+p1->id*w->vertex_size;
     } else {
     x1 = round(p1->sx); x2 = round(p2->sx);
     y1 = round(p1->sy); y2 = round(p2->sy);
     id = p1->id+p2->id*w->vertex_size;
     }

  dx = x2 - x1;
  dy = y2 - y1;

  if (abs(dx)>abs(dy)) npix = round(abs(dx));
     else npix = round(abs(dy));

  if (npix==0 || (y1<0 && y2<0) || (x1<0 && x2<0) ||
     (x1>=xres && x2>=xres) || (y1>=yres && y2>=yres)) return;

  dx /= npix;
  dy /= npix;

  x = x1;
  y = y1;

  npix++;

  while (npix-->0) {
        ix=round(x);
        iy=round(y);
        if (ix>=0 && ix<xres && iy>=0 && iy<yres)
           w->clipping_buffer[ix+iy*xres]=id;
        x+=dx;
        y+=dy;
        }

}

LIST *clip_line(WORLD *w, VERTEX *p1, VERTEX *p2)
{
  LIST *list=NULL,*new=NULL;
  float ox,oy,x,y,x1,x2,y1,y2,dx,dy;
  int ix,iy,visible=0,id,npix;

  if (p1->id>p2->id) {
     x2 = round(p1->sx); x1 = round(p2->sx);
     y2 = round(p1->sy); y1 = round(p2->sy);
     id = p2->id+p1->id*w->vertex_size;
     } else {
     x1 = round(p1->sx); x2 = round(p2->sx);
     y1 = round(p1->sy); y2 = round(p2->sy);
     id = p1->id+p2->id*w->vertex_size;
     }

  dx = x2 - x1;
  dy = y2 - y1;

  if (abs(dx)>abs(dy)) npix = round(abs(dx));
     else npix = round(abs(dy));

  if (npix==0 || (y1<0 && y2<0) || (x1<0 && x2<0) ||
     (x1>=xres && x2>=xres) || (y1>=yres && y2>=yres)) return NULL;

  dx /= npix;
  dy /= npix;

  x = x1;
  y = y1;

  npix++;

  while (npix-->0) {
        ox=x;oy=y;
        ix=round(x);
        iy=round(y);
        if (ix>=0 && ix<xres && iy>=0 && iy<yres) {
           if (w->clipping_buffer[ix+iy*xres]==id) {
              if (!visible) {
                 visible=1;
                 new=(LIST *)malloc(sizeof(LIST));
                 new->visible=1;
                 new->next=NULL;
                 new->line.v1.sx=x;
                 new->line.v1.sy=y;
                 }
              } else {
              if (visible) {
                 visible=0;
                 new->line.v2.sx=ox;
                 new->line.v2.sy=oy;
                 if (list) {
                    new->next=list;
                    list=new;
                    } else list=new;
                 }
              }
           } else {
           if (visible) {
              visible=0;
              new->line.v2.sx=ox;
              new->line.v2.sy=oy;
              if (list) {
                 new->next=list;
                 list=new;
                 } else list=new;
              }
           }
        x+=dx;
        y+=dy;
        }

  if (visible) {
     new->line.v2.sx=x2;
     new->line.v2.sy=y2;
     if (list) {
        new->next=list;
        list=new;
        } else list=new;
     }

 return list;

}

void rasterize_flat(VERTEX *p1, VERTEX *p2, VERTEX *p3, int *clip_buffer)
{

float j,d,dx1,dx2,dy1,dy2,y1,y2,y3,x1,x2,x3,y_1,y_2,y_3,dxdy,
      s,dyl,dyr,dym,dxdyr=0,dxdyl=0,xl=0,xr=0,xm=0,dxdym=0;



  x1 = p1->sx; x2 = p2->sx; x3 = p3->sx;
  y1 = p1->sy; y2 = p2->sy; y3 = p3->sy;

  if (y1>y2) {
     j=y2;y2=y1;y1=j;
     j=x2;x2=x1;x1=j;
     }

  if (y2>y3) {
     j=y3;y3=y2;y2=j;
     j=x3;x3=x2;x2=j;
     }

  if (y1>y2) {
     j=y2;y2=y1;y1=j;
     j=x2;x2=x1;x1=j;
     }

  dx1=x2-x1;dx2=x1-x3;
  dy1=y1-y2;dy2=y3-y1;

  d=dx1*dy2-dy1*dx2;
  if (d==0) return; else d=1/d;

  dx1*=d;dy1*=d;dx2*=d;dy2*=d;

  y_1 = ceil(y1);
  y_2 = ceil(y2);
  y_3 = ceil(y3);

  if (y_3-y_1==0) return;

  dyr=y_2-y_1;

  if (dyr!=0) {
     dxdy=(x2-x1)/(y2-y1);
     xr=x1+(y_1-y1)*dxdy;
     dxdyr=dxdy;
     }

  dyl=y_3-y_1;

  if (dyl!=0) {
     dxdy=(x3-x1)/(y3-y1);
     xl=x1+(y_1-y1)*dxdy;
     dxdyl=dxdy;
     }

  dym=y_3-y_2;

  if (dym!=0) {
     dxdy=(x3-x2)/(y3-y2);
     xm=x2+(y_2-y2)*dxdy;
     dxdym=dxdy;
     }

  clip_buffer+=(int)y_1*xres;

  while (dyl>0)
  {
    int x,l,r;

    if (dyr==0) {
       xr=xm;
       dxdyr=dxdym;
       dyr=dyl;
       }

    l=xl;
    r=xr;
    if (l>r) {s=l;l=r;r=s;}
    r-=l;l++;

    if (r>0 && y_1>=0 && y_1<yres && l<xres)
       for (x=l;r>0;r--,x++) if (x>=0 && x<xres) clip_buffer[x]=0;

    clip_buffer+=xres;
    xl+=dxdyl;
    xr+=dxdyr;
    y_1++;
    dyl--;
    dyr--;
  }
}


int select_line(WORLD *w, int id1, int id2)
{
 int i,key,left=0,right=w->face_size*3-1;

 if (id1>id2) {i=id1;id1=id2;id2=i;}

 key=id1+id2*w->vertex_size;

 while (left<right) {
       i=(left+right)/2;
       if (right-left==1) i=left=right;
       if (w->info[i].sortkey==key) {
          if (w->info[i].visible) return 2;
          w->info[i].visible=1;
          w->info[i].v1->used=1;
          w->info[i].v2->used=1;
          return 1;
          }
       if (w->info[i].sortkey<key) left=i; else right=i;

       }
 return 0;
}

void rasterize_face(WORLD *w, FACE *f)
{
 if (hide_removal)
    rasterize_flat(f->vertex[0],f->vertex[1],f->vertex[2],w->clipping_buffer);
 rasterize_line(w,f->vertex[0],f->vertex[1]);
 rasterize_line(w,f->vertex[1],f->vertex[2]);
 rasterize_line(w,f->vertex[2],f->vertex[0]);

}

void clipping(WORLD *w)
{
 int i;

 for (i=0;i<w->face_size*3;i++)
     if (w->info[i].visible)
        w->info[i].cut=clip_line(w,w->info[i].v1,w->info[i].v2);
}

void optimize(WORLD *w, TREE *t)
{
 if (t->more) optimize(w,t->more);

 select_line(w,t->face->vertex[0]->id,t->face->vertex[1]->id);
 select_line(w,t->face->vertex[1]->id,t->face->vertex[2]->id);
 select_line(w,t->face->vertex[2]->id,t->face->vertex[0]->id);

 if (t->less) optimize(w,t->less);
}

void rasterize(WORLD *w, TREE *t)
{
 if (t->more) rasterize(w,t->more);
 rasterize_face(w,t->face);
 if (t->less) rasterize(w,t->less);
}

void inline normalize(VECTOR v)
{
  float s;
  s = v[0]*v[0] + v[1]*v[1] + v[2]*v[2];
  s = sqrt(s);
  v[0]/=s;
  v[1]/=s;
  v[2]/=s;
}

void inline cross_product(VECTOR res, VECTOR a, VECTOR b)
{
  res[0] = b[1]*a[2]-a[1]*b[2];
  res[1] = b[2]*a[0]-a[2]*b[0];
  res[2] = b[0]*a[1]-a[0]*b[1];
}

void inline init_matrix(MATRIX *m)
{
  int i,j;
  for (j=0; j<4; j++) for (i=0; i<4; i++) (*m)[i][j] = (i==j);
}

void invert_matrix(MATRIX *s, MATRIX *d)
{
  int i,j;
  float det;

  (*d)[0][0] =   (*s)[1][1] * (*s)[2][2] - (*s)[1][2] * (*s)[2][1];
  (*d)[0][1] = - (*s)[0][1] * (*s)[2][2] + (*s)[0][2] * (*s)[2][1];
  (*d)[0][2] =   (*s)[0][1] * (*s)[1][2] - (*s)[0][2] * (*s)[1][1];

  det = (*d)[0][0] * (*s)[0][0] + (*d)[0][1] * (*s)[1][0] + (*d)[0][2] * (*s)[2][0];

  (*d)[1][0] = - (*s)[1][0] * (*s)[2][2] + (*s)[1][2] * (*s)[2][0];
  (*d)[1][1] =   (*s)[0][0] * (*s)[2][2] - (*s)[0][2] * (*s)[2][0];
  (*d)[1][2] = - (*s)[0][0] * (*s)[1][2] + (*s)[0][2] * (*s)[1][0];

  (*d)[2][0] =   (*s)[1][0] * (*s)[2][1] - (*s)[1][1] * (*s)[2][0];
  (*d)[2][1] = - (*s)[0][0] * (*s)[2][1] + (*s)[0][1] * (*s)[2][0];
  (*d)[2][2] =   (*s)[0][0] * (*s)[1][1] - (*s)[0][1] * (*s)[1][0];

  for (j=0;j<3;j++)
      for (i=0;i<3;i++) (*d)[j][i]/=det;

  (*d)[3][0] = - (*s)[3][0]*(*d)[0][0] - (*s)[3][1]*(*d)[1][0] - (*s)[3][2]*(*d)[2][0];
  (*d)[3][1] = - (*s)[3][0]*(*d)[0][1] - (*s)[3][1]*(*d)[1][1] - (*s)[3][2]*(*d)[2][1];
  (*d)[3][2] = - (*s)[3][0]*(*d)[0][2] - (*s)[3][1]*(*d)[1][2] - (*s)[3][2]*(*d)[2][2];

  (*d)[0][3] = 0;
  (*d)[1][3] = 0;
  (*d)[2][3] = 0;
  (*d)[3][3] = 1;
}

void transform(WORLD *w)
{
  float x,y,z,v;
  int i;
  MATRIX m;

  invert_matrix(w->camera,&m);

  for (i=0;i<w->vertex_size;i++)
  {
    x = w->vertex[i].x;
    y = w->vertex[i].y;
    z = w->vertex[i].z;
    w->vertex[i].tx = x*m[0][0] + y*m[1][0] + z*m[2][0] + m[3][0];
    w->vertex[i].ty = x*m[0][1] + y*m[1][1] + z*m[2][1] + m[3][1];
    w->vertex[i].tz = x*m[0][2] + y*m[1][2] + z*m[2][2] + m[3][2];
    w->vertex[i].sx = xres/2 - (fovx*w->vertex[i].tx)/w->vertex[i].tz;
    w->vertex[i].sy = yres/2 + (fovy*w->vertex[i].ty)/w->vertex[i].tz;
  }

  for (i=0;i<w->face_size;i++) {

      v=(w->face[i].vertex[1]->sy - w->face[i].vertex[0]->sy)*
        (w->face[i].vertex[2]->sx - w->face[i].vertex[0]->sx)-
        (w->face[i].vertex[2]->sy - w->face[i].vertex[0]->sy)*
        (w->face[i].vertex[1]->sx - w->face[i].vertex[0]->sx);

      v=invert_normals ? -v : v;

      if (v>0 || !hide_removal) {

         TREE *leaf,*new_leaf;

         new_leaf=(TREE *)malloc(sizeof(TREE));
         new_leaf->z = w->face[i].vertex[0]->tz+
                       w->face[i].vertex[1]->tz+
                       w->face[i].vertex[2]->tz;
         new_leaf->face = &w->face[i];
         new_leaf->more = NULL;
         new_leaf->less = NULL;

         if (!w->tree) w->tree=new_leaf; else {
            leaf=w->tree;
            while (leaf!=NULL) {
                  if (leaf->z < new_leaf->z) {
                     if (leaf->more) leaf=leaf->more; else {
                        leaf->more=new_leaf;
                        break;
                        }
                     } else {
                     if (leaf->less) leaf=leaf->less; else {
                        leaf->less=new_leaf;
                        break;
                        }
                     }
                  }
            }
         }
      }
}

MATRIX *create_camera(float tx, float ty, float tz,
                      float px, float py, float pz)
{
  MATRIX *camera;
  VECTOR view_dir,right,up,world_up;
  camera=(MATRIX *)malloc(sizeof(MATRIX));

  world_up[0] = 0;
  world_up[1] = 1;
  world_up[2] = 0;

  view_dir[0] = (tx-px);
  view_dir[1] = (ty-py);
  view_dir[2] = (tz-pz);

  normalize(view_dir);
  cross_product(right,world_up,view_dir);
  cross_product(up,right,view_dir);
  normalize(up);
  normalize(right);

  init_matrix(camera);
  (*camera)[0][0]=right[0];
  (*camera)[0][1]=right[1];
  (*camera)[0][2]=right[2];
  (*camera)[1][0]=up[0];
  (*camera)[1][1]=up[1];
  (*camera)[1][2]=up[2];
  (*camera)[2][0]=view_dir[0];
  (*camera)[2][1]=view_dir[1];
  (*camera)[2][2]=view_dir[2];
  (*camera)[3][0]=px;
  (*camera)[3][1]=py;
  (*camera)[3][2]=pz;

  return camera;
}

void sort_info(WORLD *w, int left, int right)
{

 INFO swp;
 int start=left,
     stop=right,
     middle=w->info[left].sortkey;

 while(1) {

 while (w->info[left].sortkey<middle) left++;
 while (w->info[right].sortkey>middle) right--;

 if (left>right) break;

 memcpy(&swp,&w->info[left],sizeof(INFO));
 memcpy(&w->info[left++],&w->info[right],sizeof(INFO));
 memcpy(&w->info[right--],&swp,sizeof(INFO));

 }

 if (left<stop) sort_info(w,left,stop);
 if (right>start) sort_info(w,start,right);

}

WORLD *load_asc(char *name)
{
 FILE *f;
 WORLD *w;
 char s[MAX_INPUT_LINE_LENGTH];
 int i,vertex_id=0,face_id=0,new_id=0,old_id=0,hash_base=0,oldkey=0;

 f=fopen(name,"r");
 w=(WORLD *)malloc(sizeof(WORLD));
 w->vertex_size=0;
 w->face_size=0;
 w->tree=NULL;

 w->clipping_buffer=(int *)malloc(sizeof(int)*xres*yres);
 memset(w->clipping_buffer,0,sizeof(int)*xres*yres);

 while (!feof(f)) {
  fgets(s,MAX_INPUT_LINE_LENGTH,f);
  if (strstr(s,"Vertex ") && !strstr(s,"Vertex list"))
     w->vertex_size++;
  if (strstr(s,"Face ") && !strstr(s,"Face list"))
     w->face_size++;
  if (strstr(s,"Camera ")) {
     float fov,xpos,ypos,zpos,xtar,ytar,ztar,
           bank_angle,per_near,per_far;
     sscanf(s,"Camera (%fmm)",&fov);
     fgets(s,MAX_INPUT_LINE_LENGTH,f);
     sscanf(s,"Position:  X:%f Y:%f Z:%f",&xpos,&zpos,&ypos);
     fgets(s,MAX_INPUT_LINE_LENGTH,f);
     sscanf(s,"Target:  X:%f Y:%f Z:%f",&xtar,&ztar,&ytar);
     fgets(s,MAX_INPUT_LINE_LENGTH,f);
     sscanf(s,"Bank angle: %f degrees",&bank_angle);
     fgets(s,MAX_INPUT_LINE_LENGTH,f);
     sscanf(s,"Near %f Far %f",&per_near,&per_far);
     w->camera=create_camera(xtar,ytar,ztar,xpos,ypos,zpos);
     }
  }

 fclose(f);

 w->vertex=(VERTEX *)malloc(w->vertex_size*sizeof(VERTEX));
 w->face=(FACE *)malloc(w->face_size*sizeof(FACE));
 w->info=(INFO *)malloc(w->face_size*sizeof(INFO)*3);

 f=fopen(name,"r");

 while (!feof(f)) {
  fgets(s,MAX_INPUT_LINE_LENGTH,f);
  if (strstr(s,"Vertex ") && !strstr(s,"Vertex list")) {
     w->vertex[vertex_id].id=vertex_id;
     w->vertex[vertex_id].used=0;
     sscanf(s,"Vertex %d:  X: %f     Y: %f     Z: %f",&new_id,
     &w->vertex[vertex_id].x,&w->vertex[vertex_id].z,&w->vertex[vertex_id].y);
     if (new_id<old_id) hash_base=vertex_id;
     old_id=new_id;
     vertex_id++;
     }
  if (strstr(s,"Face ") && !strstr(s,"Face list")) {
     int id0,id1,id2,ab,bc,ca;
     sscanf(s,"Face %d:    A:%d B:%d C:%d AB:%d BC:%d CA:%d",&new_id,
     &id0,&id1,&id2,&ab,&bc,&ca);
     w->face[face_id].vertex[0]=&w->vertex[id0+hash_base];
     w->face[face_id].vertex[1]=&w->vertex[id1+hash_base];
     w->face[face_id].vertex[2]=&w->vertex[id2+hash_base];
     face_id++;
     }
  }

 fclose(f);

 for (i=0;i<w->face_size;i++) {
     w->info[i*3].v1_id=w->face[i].vertex[0]->id;
     w->info[i*3].v2_id=w->face[i].vertex[1]->id;
     w->info[i*3].v1=w->face[i].vertex[0];
     w->info[i*3].v2=w->face[i].vertex[1];
     w->info[i*3].visible=0;
     w->info[i*3+1].v1_id=w->face[i].vertex[1]->id;
     w->info[i*3+1].v2_id=w->face[i].vertex[2]->id;
     w->info[i*3+1].v1=w->face[i].vertex[1];
     w->info[i*3+1].v2=w->face[i].vertex[2];
     w->info[i*3+1].visible=0;
     w->info[i*3+2].v1_id=w->face[i].vertex[2]->id;
     w->info[i*3+2].v2_id=w->face[i].vertex[0]->id;
     w->info[i*3+2].v1=w->face[i].vertex[2];
     w->info[i*3+2].v2=w->face[i].vertex[0];
     w->info[i*3+2].visible=0;
     }

 for (i=0;i<w->face_size*3;i++) {
     if (w->info[i].v1_id>w->info[i].v2_id) {
        int v=w->info[i].v2_id;
        w->info[i].v2_id=w->info[i].v1_id;
        w->info[i].v1_id=v;
        }
     w->info[i].sortkey=w->info[i].v1_id+w->info[i].v2_id*w->vertex_size;
     }

 sort_info(w,0,w->face_size*3-1);

 for (i=0;i<w->face_size*3;i++)
     if (w->info[i].sortkey==oldkey) w->info[i].sortkey=0; else
     oldkey=w->info[i].sortkey;

 sort_info(w,0,w->face_size*3-1);

 return w;

}

void kill_tree(TREE *t)
{
 if (t->more) kill_tree(t->more);
 if (t->less) kill_tree(t->less);
 free(t);
}

void kill(WORLD *w)
{

 int i;

 kill_tree(w->tree);

 for (i=0;i<w->face_size*3;i++)
     if (w->info[i].visible) {
        LIST *_c,*c=w->info[i].cut;
        while (c) {
              _c=c;
              c=c->next;
              free(_c);
              }
        }

 free(w->clipping_buffer);
 free(w->camera);
 free(w->vertex);
 free(w->face);
 free(w->info);
 free(w);

}

void save_mf(char *name, WORLD *w, int id, int last_id)
{
  FILE *f;
  int i;

  if (id==1) {
     f=fopen(name,"w");
     fprintf(f,"mode_setup;\nunit#=%s#;\nmypen#=%s#;\n"
               "define_pixels(unit,mypen);\n\n",hsize,penth);
     } else f=fopen(name,"a");

  fprintf(f,"beginchar(%d,unit#,%sunit#,0);\n"
  "pickup pencircle scaled mypen;\n"
  "def myline(expr a,b)= draw a--b; enddef;\n",id,aspect);

  for (i=0;i<w->face_size*3;i++)
      if (w->info[i].visible) {
         LIST *c=w->info[i].cut;
            while (c) {
             fprintf(f,"myline((w*%f,h*%f),(w*%f,h*%f));\n",
             c->line.v1.sx/xres,(yres-c->line.v1.sy)/yres,
             c->line.v2.sx/xres,(yres-c->line.v2.sy)/yres);
             c=c->next;
             }
         }

  if (border) {
     fprintf(f,"myline((w*0,h*0),(w*1,h*0));\n");
     fprintf(f,"myline((w*1,h*0),(w*1,h*1));\n");
     fprintf(f,"myline((w*1,h*1),(w*0,h*1));\n");
     fprintf(f,"myline((w*0,h*1),(w*0,h*0));\n");
     }

  fprintf(f,"endchar;\n\n");
  if (id==last_id) fprintf(f,"end;\n");

  fclose(f);
}

void convert(char *asc, char *mf, int id, int last_id)
{

 WORLD *world;

 printf(" [] Loading (%s)\n",asc);

 world=load_asc(asc);

 printf(" [] Optimizing MF code\n");

 transform(world);
 optimize(world,world->tree);
 rasterize(world,world->tree);
 clipping(world);

 printf(" [] Saving (%s)\n",mf);

 save_mf(mf,world,id,last_id);

 kill(world);

}

int main(int argc, char **argv)
{
 int output_id=1,i;

 printf(" [] This is Meta3D v0.5 by ReDox/MovSD (a.k.a. Daniel S\\'{y}kora) * 1999\n\n");

 if (argc<3) {
    printf(" Usage: meta3d <input.asc> <output.mf> [<param...>]\n"
    "        input.asc    : standard 3D Studio text output format\n"
    "                       important sections: Face, Vertex and Camera\n"
    "                       other sections and information are invisible\n"
    "                       to convert all *.asc files in dir to one output.mf\n"
    "                       as a seqvenece of independent pictures use asterix\n"
    "     -h --hsize n    : horizontal size of picture in MF units (=10cm)\n"
    "     -a --aspect n   : float number as picture aspect ratio (=0.625)\n"
    "     -p --penth n    : pen thickness for lines in MF units (=0.1mm)\n"
    "     -r --raster x y : size of rasterization buffer (=1280x800 (!) 4MB)\n"
    "     -n --normals    : invert normals to compute corrcet hide-removal\n"
    "     -f --fov x y    : field of view coeficients (=1024x1024)\n"
    "     -i --hiderem    : turn off hide-removal rasterization\n"
    "     -b --border     : create visible picture border\n\n"
    " see readme.txt for more information and some examples\n");
    return 1;
    }

 while (output_id<argc)
       if (strstr(argv[output_id],"-")==NULL)
          output_id++; else break;

 i=--output_id;

 while (++i<argc) while (1) {
       if (strstr(argv[i],"-h")) {
          hsize=argv[++i];
          break;
          }
       if (strstr(argv[i],"-a")) {
          aspect=argv[++i];
          break;
          }
       if (strstr(argv[i],"-p")) {
          penth=argv[++i];
          break;
          }
       if (strstr(argv[i],"-r")) {
          xres=atoi(argv[++i]);
          yres=atoi(argv[++i]);
          break;
          }
       if (strstr(argv[i],"-n")) {
          invert_normals=1;
          break;
          }
       if (strstr(argv[i],"-f")) {
          fovx=atoi(argv[++i]);
          fovy=atoi(argv[++i]);
          break;
          }
       if (strstr(argv[i],"-i")||
           strstr(argv[i],"-hi")) {
          hide_removal=0;
          break;
          }
       if (strstr(argv[i],"-b")) {
          border=1;
          break;
          }
       break;
       }

 for (i=1;i<output_id;i++)
     convert(argv[i],argv[output_id],i,output_id-1);

 return 1;

}
